/* -*- Mode: C; c-basic-offset:4 ; -*- */
/*
*  (C) 2001 by Argonne National Laboratory.
*      See COPYRIGHT in top-level directory.
*/

#include "mpidi_ch3_impl.h"

/*
* MPIDI_CH3_iRead()
*/
#undef FUNCNAME
#define FUNCNAME MPIDI_CH3_iRead
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
void MPIDI_CH3_iRead(MPIDI_VC * vc, MPID_Request * rreq)
{
    void *mem_ptr;
    char *iter_ptr;
    int num_bytes;
    MPIDI_CH3I_SHM_Packet_t *pkt_ptr;
    register int index;
    int cur_index = 0;
    MPIDI_STATE_DECL(MPID_STATE_MEMCPY);
    MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3_IREAD);

    MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3_IREAD);
    MPIDI_DBG_PRINTF((60, FCNAME, "ch3_iread\n"));

    if (vc->mm.bShm)
    {
	index = vc->mm.read_shmq->head_index;
	if (vc->mm.read_shmq->packet[index].avail == MPIDI_CH3I_PKT_AVAILABLE)
	{
	    rreq->mm.iov_offset = 0;
	    vc->mm.recv_active = rreq;
	    MPIDI_CH3I_SHM_post_readv(vc, rreq->ch3.iov + rreq->mm.iov_offset, rreq->ch3.iov_count - rreq->mm.iov_offset, NULL);
	    return;
	}

	mem_ptr = (void*)vc->mm.read_shmq->packet[index].cur_pos;
	pkt_ptr = &vc->mm.read_shmq->packet[index];
	num_bytes = vc->mm.read_shmq->packet[index].num_bytes;
	assert(num_bytes > 0);

	iter_ptr = mem_ptr;

	rreq->mm.iov_offset = 0;
	while (rreq->mm.iov_offset < rreq->ch3.iov_count)
	{
	    if ((int)rreq->ch3.iov[rreq->mm.iov_offset].MPID_IOV_LEN <= num_bytes)
	    {
		MPIDI_FUNC_ENTER(MPID_STATE_MEMCPY);
		memcpy(rreq->ch3.iov[rreq->mm.iov_offset].MPID_IOV_BUF, iter_ptr, rreq->ch3.iov[rreq->mm.iov_offset].MPID_IOV_LEN);
		MPIDI_FUNC_EXIT(MPID_STATE_MEMCPY);
		iter_ptr += rreq->ch3.iov[rreq->mm.iov_offset].MPID_IOV_LEN;
		num_bytes -= rreq->ch3.iov[rreq->mm.iov_offset].MPID_IOV_LEN;
		rreq->mm.iov_offset += 1;
	    }
	    else
	    {
		MPIDI_FUNC_ENTER(MPID_STATE_MEMCPY);
		memcpy(rreq->ch3.iov[rreq->mm.iov_offset].MPID_IOV_BUF, iter_ptr, num_bytes);
		MPIDI_FUNC_EXIT(MPID_STATE_MEMCPY);

		pkt_ptr->cur_pos = pkt_ptr->data;
		pkt_ptr->avail = MPIDI_CH3I_PKT_AVAILABLE;
		vc->mm.read_shmq->head_index = (index + 1) % MPIDI_CH3I_NUM_PACKETS;

		rreq->ch3.iov[rreq->mm.iov_offset].MPID_IOV_BUF = (char *) rreq->ch3.iov[rreq->mm.iov_offset].MPID_IOV_BUF + num_bytes;
		rreq->ch3.iov[rreq->mm.iov_offset].MPID_IOV_LEN -= num_bytes;
		vc->mm.recv_active = rreq;
		MPIDI_CH3I_SHM_post_readv(vc, rreq->ch3.iov + rreq->mm.iov_offset, rreq->ch3.iov_count - rreq->mm.iov_offset, NULL);
		MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3_IREAD);
		return;
	    }
	}
	pkt_ptr->cur_pos += (pkt_ptr->num_bytes - num_bytes);
	pkt_ptr->num_bytes = num_bytes;
	if (pkt_ptr->num_bytes == 0)
	{
	    pkt_ptr->cur_pos = pkt_ptr->data;
	    pkt_ptr->avail = MPIDI_CH3I_PKT_AVAILABLE;
	    vc->mm.read_shmq->head_index = (index + 1) % MPIDI_CH3I_NUM_PACKETS;
	}

	if (rreq->ch3.ca == MPIDI_CH3_CA_COMPLETE)
	{
	    /* mark data transfer as complete and decrement CC */
	    MPIDI_CH3U_Request_complete(rreq);
	}
	else
	{
	    /* FIXME: excessive recursion... */
	    MPIDI_CH3U_Handle_recv_req(vc, rreq);
	}

    }
    else
    {
	assert(vc->mm.state == MPIDI_CH3I_VC_STATE_CONNECTED);
	rreq->mm.iov_offset = 0;

	MPIDI_CH3I_MM_VC_post_read(vc, rreq);
    }

    MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3_IREAD);
}
